@@ -17,15 +17,15 @@
 */
 
 #include "sh.h"
-#ifdef HAVE_SETRLIMIT
-# include <sys/time.h>
+#include "ksh_time.h"
+#ifdef HAVE_SYS_RESOURCE_H
 # include <sys/resource.h>
-#endif /* HAVE_SETRLIMIT */
+#endif /* HAVE_SYS_RESOURCE_H */
 #ifdef HAVE_ULIMIT_H
 # include <ulimit.h>
 #else /* HAVE_ULIMIT_H */
 # ifdef HAVE_ULIMIT
-extern	long ulimit ARGS((int, ...));
+extern	long ulimit();
 # endif /* HAVE_ULIMIT */
 #endif /* HAVE_ULIMIT_H */
 
@@ -159,7 +159,13 @@ c_ulimit(wp)
 			val = RLIM_INFINITY;
 		else
 #endif /* RLIM_INFINITY */
-			val = evaluate(p) * l->factor;
+		{
+			long rval;
+
+			if (!evaluate(p, &rval, TRUE))
+				return 1;
+			val = rval * l->factor;
+		}
 	}
 	if (all) {
 		for (l = limits; l->name; l++) {
