@@ -3,6 +3,7 @@
  */
 
 #include "sh.h"
+#include "c_test.h"
 #include <ctype.h>
 #include "ksh_stat.h"
 
@@ -15,7 +16,14 @@ static int	herein ARGS((char *hname, int sub));
 #ifdef KSH
 static char 	*do_selectargs ARGS((char **ap));
 static int	pr_menu ARGS((register char **ap));
-#endif KSH
+#endif /* KSH */
+#ifdef KSH
+static int	dbteste_isa ARGS((Test_env *te, Test_meta meta));
+static char	*dbteste_getopnd ARGS((Test_env *te, Test_op op, int do_eval));
+static int	dbteste_eval ARGS((Test_env *te, Test_op op, char *opnd1,
+				char *opnd2, int do_eval));
+static void	dbteste_error ARGS((Test_env *te, int offset, char *msg));
+#endif /* KSH */
 
 
 /*
@@ -83,8 +91,7 @@ execute(t, flags)
 		/* POSIX says expand command words first, then redirections,
 		 * and assignments last..
 		 */
-		ap = eval(t->args, t->evalflags ? t->evalflags
-						: DOBLANK|DOGLOB|DOTILDE);
+		ap = eval(t->args, t->evalflags | DOBLANK | DOGLOB | DOTILDE);
 		if (Flag(FXTRACE) && ap[0]) {
 			shf_fprintf(shl_out, "%s",
 				substitute(strval(global("PS4")), 0));
@@ -203,7 +210,11 @@ execute(t, flags)
 	  }
 
 	  case TASYNC:
-		rv = execute(t->left, flags|XBGND|XFORK);
+		/* XXX non-optimal, I think - "(foo &)", forks for (),
+		 * forks again for async...  parent should optimize
+		 * this to "foo &"...
+		 */
+		rv = execute(t->left, (flags&~XEXEC)|XBGND|XFORK);
 		break;
 
 	  case TOR:
@@ -222,19 +233,16 @@ execute(t, flags)
 #ifdef KSH
 	  case TDBRACKET:
 	    {
-		XPtrV ap;
+		Test_env te;
 
-		/* Must build arguments one by one because the second
-		 * operand of = and != need to be eval'd with DOPAT.
-		 */
-		for (i = 0; t->args[i]; i++)
-			;
-		XPinit(ap, i + 1);
-		for (i = 0; t->args[i]; i++)
-			XPput(ap, evalstr(t->args[i], DOTILDE
-				    |(t->vars[i][1] == DB_PAT ? DOPAT : 0)));
-		XPput(ap, (char *) 0);
-		rv = call_builtin(findcom("test", FC_BI), (char **) XPptrv(ap));
+		te.flags = TEF_DBRACKET;
+		te.pos.wp = t->args;
+		te.isa = dbteste_isa;
+		te.getopnd = dbteste_getopnd;
+		te.eval = dbteste_eval;
+		te.error = dbteste_error;
+
+		rv = test_parse(&te);
 		break;
 	    }
 #endif /* KSH */
@@ -319,7 +327,8 @@ execute(t, flags)
 		cp = evalstr(t->str, DOTILDE);
 		for (t = t->left; t != NULL && t->type == TPAT; t = t->right)
 		    for (ap = t->vars; *ap; ap++)
-			if ((s = evalstr(*ap, DOTILDE|DOPAT)) && gmatch(cp, s))
+			if ((s = evalstr(*ap, DOTILDE|DOPAT))
+			    && gmatch(cp, s, FALSE))
 				goto Found;
 		break;
 	  Found:
@@ -388,7 +397,7 @@ comexec(t, tp, ap, flags)
 	register char **lastp;
 	static struct op texec = {TEXEC};
 	int type_flags;
-	int no_keepasn;
+	int keepasn_ok;
 	int fcflags = FC_BI|FC_FUNC|FC_PATH;
 
 	/* snag the last argument for $_ XXX not the same as at&t ksh,
@@ -414,7 +423,7 @@ comexec(t, tp, ap, flags)
 	 *   FOO=bar command			FOO is neither kept nor exported
 	 *   PATH=... foobar			use new PATH in foobar search
 	 */
-	no_keepasn = 0;
+	keepasn_ok = 1;
 	while (tp && tp->type == CSHELL) {
 		fcflags = FC_BI|FC_FUNC|FC_PATH;/* undo effects of command */
 		if (tp->val.f == c_builtin) {
@@ -448,7 +457,7 @@ comexec(t, tp, ap, flags)
 			if (saw_p) {
 				if (Flag(FRESTRICTED)) {
 					warningf(TRUE,
-						"command -p: restricted\n");
+						"command -p: restricted");
 					rv = 1;
 					goto Leave;
 				}
@@ -458,7 +467,7 @@ comexec(t, tp, ap, flags)
 			/* POSIX says special builtins loose their status
 			 * if accessed using command.
 			 */
-			no_keepasn = 1;
+			keepasn_ok = 0;
 			if (!ap[0]) {
 				/* ensure command with no args exits with 0 */
 				subst_exstat = 0;
@@ -471,13 +480,15 @@ comexec(t, tp, ap, flags)
 	/* todo: POSIX says assignments preceding a function are kept, at&t
 	 * ksh does not do this
 	 */
-	if (!no_keepasn && (!ap[0] || (tp && tp->flag & KEEPASN)))
+	if (keepasn_ok && (!ap[0] || (tp && tp->flag & KEEPASN)))
 		type_flags = 0;
 	else {
 		/* create new variable/function block */
 		newblock();
 		type_flags = LOCAL|EXPORT;
 	}
+	if (Flag(FEXPORT))
+		type_flags |= EXPORT;
 	for (i = 0; t->vars[i]; i++) {
 		cp = evalstr(t->vars[i], DOASNTILDE);
 		if (Flag(FXTRACE)) {
@@ -497,7 +508,7 @@ comexec(t, tp, ap, flags)
 		goto Leave;
 	} else if (!tp) {
 		if (Flag(FRESTRICTED) && strchr_dirsep(cp)) {
-			warningf(TRUE, "%s: restricted\n", cp);
+			warningf(TRUE, "%s: restricted", cp);
 			rv = 1;
 			goto Leave;
 		}
@@ -513,6 +524,7 @@ comexec(t, tp, ap, flags)
 	  {
 		volatile int old_xflag;
 		volatile int old_inuse;
+		char *volatile old_kshname;
 
 		if (!(tp->flag & ISSET)) {
 			struct tbl *ftp;
@@ -520,23 +532,22 @@ comexec(t, tp, ap, flags)
 			if (!tp->u.fpath) {
 				/* XXX: exit code 126 vs 127 */
 				warningf(TRUE,
-				"%s: can't find function definition file\n",
-					cp);
+				"%s: can't find function definition file", cp);
 				rv = 127;
 				break;
 			}
 			if (include(tp->u.fpath, 0, (char **) 0) < 0) {
 				warningf(TRUE,
-			    "%s: can't open function definition file %s - %s\n",
+			    "%s: can't open function definition file %s - %s",
 					cp, tp->u.fpath, strerror(errno));
 				rv = 126;
 				break;
 			}
-			if (!(ftp = findfunc(tp->name, hash(tp->name), FALSE))
+			if (!(ftp = findfunc(cp, hash(cp), FALSE))
 			    || !(ftp->flag & ISSET))
 			{
 				warningf(TRUE,
-					"%s: function not defined by %s\n",
+					"%s: function not defined by %s",
 					cp, tp->u.fpath);
 				rv = 127;
 				break;
@@ -544,8 +555,10 @@ comexec(t, tp, ap, flags)
 			tp = ftp;
 		}
 
-		/* XXX: posix says $0 remains unchanged */
-		kshname = ap[0];
+		/* posix says $0 remains unchanged, at&t ksh changes it */
+		old_kshname = kshname;
+		if (!Flag(FPOSIX))
+			kshname = ap[0];
 		e->loc->argv = ap;
 		for (i = 0; *ap++ != NULL; i++)
 			;
@@ -564,6 +577,7 @@ comexec(t, tp, ap, flags)
 			exstat = execute(tp->val.t, flags & XERROK);
 			i = LRETURN;
 		}
+		kshname = old_kshname;
 		Flag(FXTRACE) = old_xflag;
 		tp->flag = (tp->flag & ~FINUSE) | old_inuse;
 		/* Were we deleted while executing?  If so, free the execution
@@ -618,9 +632,9 @@ comexec(t, tp, ap, flags)
 			 * is not found.
 			 */
 			if (access(cp, F_OK) < 0)
-				warningf(TRUE, "%s: not found\n", cp);
+				warningf(TRUE, "%s: not found", cp);
 			else
-				warningf(TRUE, "%s: cannot execute\n", cp);
+				warningf(TRUE, "%s: cannot execute", cp);
 			/* XXX posix says 126 if in path and cannot execute */
 			rv = 127;
 			break;
@@ -692,7 +706,18 @@ scriptexec(tp, ap)
 
 				while (*cp && *cp != '\n' && *cp != ' '
 				       && *cp != '\t')
+				{
+#ifdef OS2
+			/* Allow shell search without prepended path.
+			 * Use / explicitly so \ can be used if explicit
+			 * path is needed.
+			 * [This could surprise people...]
+			 */
+					if (*cp == '/')
+						a0 = cp + 1;
+#endif /* OS2 */
 					cp++;
+				}
 				if (*cp && *cp != '\n') {
 					*cp++ = '\0';
 					while (*cp
@@ -974,10 +999,12 @@ search_access1(path, mode)
 	char *path;
 	int mode;
 {
-	static char *xsuffixes[] = { ".", ".ksh", ".cmd", ".exe", ".com",
-				     ".bat", (char *) 0
+	static char *xsuffixes[] = { ".ksh", ".sh", ".cmd", ".exe", ".com",
+				     ".bat", ".", (char *) 0
+				   };
+	static char *rsuffixes[] = { ".ksh", ".sh", ".cmd", ".bat", ".",
+				      (char *) 0
 				   };
-	static char *rsuffixes[] = { ".", ".ksh", ".cmd", ".bat", (char *) 0 };
 	int i;
 	char *tp = path + strlen(path);
 	char *p;
@@ -1008,7 +1035,7 @@ search_access1(path, mode)
 	sfx = mode == R_OK ? rsuffixes : xsuffixes;
 	for (i = 0; sfx[i]; i++) {
 		strcpy(tp, p = sfx[i]);
-		if ((ret = search_access(path, R_OK)) == 0)
+		if (search_access(path, R_OK) == 0)
 			return 0;
 		*tp = '\0';
 	}
@@ -1174,7 +1201,7 @@ iosetup(iop)
 				   X_OK | ((iop->flag & IORDUP) ? R_OK : W_OK),
 				   &emsg);
 			if (u < 0) {
-				warningf(TRUE, "%d>&%s: %s\n",
+				warningf(TRUE, "%d>&%s: %s",
 					iop->unit, cp, emsg);
 				return -1;
 			}
@@ -1195,7 +1222,7 @@ iosetup(iop)
 	}
 	if (do_open) {
 		if (Flag(FRESTRICTED) && (flags & O_CREAT)) {
-			warningf(TRUE, "%s: restricted\n", cp);
+			warningf(TRUE, "%s: restricted", cp);
 			return -1;
 		}
 		u = open(cp, flags, 0666);
@@ -1207,7 +1234,7 @@ iosetup(iop)
 	if (u < 0) {
 		/* herein() may already have printed message */
 		if (u == -1)
-			warningf(TRUE, "cannot %s %s: %s\n",
+			warningf(TRUE, "cannot %s %s: %s",
 			       (iop->flag&IOTYPE) == IODUP ? "dup"
 				: ((iop->flag&IOTYPE) == IOREAD
 				    || (iop->flag&IOTYPE) == IOHERE) ? "open"
@@ -1237,7 +1264,7 @@ herein(hname, sub)
 
 	/* ksh -c 'cat << EOF' can cause this... */
 	if (hname == (char *) 0) {
-		warningf(TRUE, "here document missing\n");
+		warningf(TRUE, "here document missing");
 		return -2;
 	}
 	if (sub) {
@@ -1277,7 +1304,7 @@ herein(hname, sub)
 		if (shf == NULL
 		    || (fd = open(h->name, O_RDONLY, 0)) < 0)
 			errorf("%s: %s", h->name, strerror(errno));
-		shf_write(cp, strlen(cp), shf);
+		shf_puts(cp, shf);
 		if (shf_close(shf) == EOF) {
 			close(fd);
 			shf = (struct shf *) 0;
@@ -1394,3 +1421,102 @@ pr_menu(ap)
 	return n;
 }
 #endif /* KSH */
+#ifdef KSH
+
+/*
+ *	[[ ... ]] evaluation routines
+ */
+
+extern const char *const dbtest_tokens[];
+extern const char db_close[];
+
+/* Test if the current token is a whatever.  Accepts the current token if
+ * it is.  Returns 0 if it is not, non-zero if it is (in the case of
+ * TM_UNOP and TM_BINOP, the returned value is a Test_op).
+ */
+static int
+dbteste_isa(te, meta)
+	Test_env *te;
+	Test_meta meta;
+{
+	int ret = 0;
+	int uqword;
+	char *p;
+
+	if (!*te->pos.wp)
+		return meta == TM_END;
+
+	/* unquoted word? */
+	for (p = *te->pos.wp; *p == CHAR; p += 2)
+		;
+	uqword = *p == EOS;
+
+	if (meta == TM_UNOP || meta == TM_BINOP) {
+		if (uqword) {
+			char buf[8];	/* longer than the longest operator */
+			char *q = buf;
+			for (p = *te->pos.wp; *p == CHAR
+					      && q < &buf[sizeof(buf) - 1];
+					      p += 2)
+				*q++ = p[1];
+			*q = '\0';
+			ret = (int) test_isop(te, meta, buf);
+		}
+	} else if (meta == TM_END)
+		ret = 0;
+	else
+		ret = uqword
+			&& strcmp(*te->pos.wp, dbtest_tokens[(int) meta]) == 0;
+
+	/* Accept the token? */
+	if (ret)
+		te->pos.wp++;
+
+	return ret;
+}
+
+static char *
+dbteste_getopnd(te, op, do_eval)
+	Test_env *te;
+	Test_op op;
+	int do_eval;
+{
+	char *s = *te->pos.wp;
+
+	if (!s)
+		return (char *) 0;
+
+	te->pos.wp++;
+
+	if (!do_eval)
+		return null;
+
+	if (op == TO_STGT || op == TO_STLT)
+		s = evalstr(s, DOTILDE | DOPAT);
+	else
+		s = evalstr(s, DOTILDE);
+
+	return s;
+}
+
+static int
+dbteste_eval(te, op, opnd1, opnd2, do_eval)
+	Test_env *te;
+	Test_op op;
+	char *opnd1;
+	char *opnd2;
+	int do_eval;
+{
+	return test_eval(te, op, opnd1, opnd2, do_eval);
+}
+
+static void
+dbteste_error(te, offset, msg)
+	Test_env *te;
+	int offset;
+	char *msg;
+{
+	te->flags |= TEF_ERROR;
+	internal_errorf(0, "dbteste_error: %s (offset %d)", msg, offset);
+}
+#endif /* KSH */
