@@ -35,8 +35,7 @@ struct source {
 #define	SWORDS	6		/* string[] */
 #define	SWORDSEP 7		/* string[] seperator */
 #define	SALIAS	8		/* alias expansion */
-#define	SHIST	9		/* history expansion */
-#define SREREAD	10		/* read ahead to be re-scanned */
+#define SREREAD	9		/* read ahead to be re-scanned */
 
 /* Source.flags values */
 #define SF_ECHO		BIT(0)	/* echo niput to shlout */
@@ -120,5 +119,4 @@ EXTERN	char	ident [IDENT+1];
 EXTERN	char  **history;	/* saved commands */
 EXTERN	char  **histptr;	/* last history item */
 EXTERN	int	histsize;	/* history size */
-EXTERN	int	histpush;	/* number of pushed fc commands */
 #endif /* HISTORY */
