@@ -23,7 +23,7 @@ static	int	is_restricted ARGS((char *name));
 static	char	initifs [] = "IFS= \t\n"; /* must be R/W */
 
 static	const	char   initsubs [] = 
-  "${PS1=$ } ${PS2=> } ${PS3=#? } ${PS4=+ }";
+  "${PS2=> } ${PS3=#? } ${PS4=+ }";
 
 static const char version_param[] =
 #ifdef KSH
@@ -35,7 +35,7 @@ static const char version_param[] =
 
 static	const	char *initcoms [] = {
 	"typeset", "-x", "SHELL", "PATH", "HOME", NULL,
-	"typeset", "-r", version_param, "PWD", "OLDPWD", NULL,
+	"typeset", "-r", version_param, NULL,
 	"typeset", "-ri", "PPID", NULL,
 	"typeset", "-i", "OPTIND=1", "MAILCHECK=600",
 #ifdef KSH
@@ -61,7 +61,6 @@ static	const	char *initcoms [] = {
 #endif /* KSH */
 #ifdef KSH
 	 /* Aliases that are builtin commands in at&t */
-	  "pwd=print -r - \"$PWD\"",
 	  "login=exec login",
 	  "newgrp=exec newgrp",
 #endif /* KSH */
@@ -81,14 +80,11 @@ main(argc, argv, envp)
 	char **envp;
 {
 	register int i;
-	char *arg;
 	int argi;
 	Source *s;
 	struct block *l;
 	int restricted;
 	char **wp;
-	struct stat s1, s2;
-	extern char ksh_version [];
 	struct env env;
 
 #ifdef MEM_DEBUG
@@ -158,6 +154,13 @@ main(argc, argv, envp)
 	path = def_path;
 
 
+	/* Turn on nohup by default for how - will change to off
+	 * by default once people are aware of its existance
+	 * (at&t ksh does not have a nohup option - it always sends
+	 * the hup).
+	 */
+	Flag(FNOHUP) = 1;
+
 	/* Turn on brace expansion by default.  At&t ksh's that have
 	 * alternation always have it on.  BUT, posix doesn't have
 	 * brace expansion, so set this before setting up FPOSIX
@@ -188,29 +191,28 @@ main(argc, argv, envp)
 	/* assign default shell variable values */
 	substitute(initsubs, 0);
 
-	/* assign PWD (using enviroment PWD if it passes test below) */
-	arg = strval(global("PWD"));
-	if (*arg == '\0'
-		|| stat(arg, &s1) < 0 || stat(".", &s2) < 0
-		|| s1.st_dev != s2.st_dev || s1.st_ino != s2.st_ino)
+	/* Figure out the current working directory and set $PWD */
 	{
-		/* Ignore imported PWD - it's wrong */
-		arg = alloc(PATH, ATEMP);
-		if (getcwd(arg, PATH) == (char *) 0) {
-			warningf(FALSE,
-		    "Can't find current directory (%s), changing to /\n",
-				strerror(errno));
-			if (arg)
-				afree(arg, ATEMP);
-			arg = slash;
-			(void) chdir(arg);
-		}
-	} else
-		/* Dup arg so we don't copy over its old value */
-		arg = strsave(arg, ATEMP);
-	setstr(global("PWD"), arg);
-	if (arg != slash)
-		afree(arg, ATEMP);
+		struct stat s_pwd, s_dot;
+		struct tbl *pwd_v = global("PWD");
+		char *pwd = strval(pwd_v);
+		char *pwdx = pwd;
+
+		/* Try to use existing $PWD if it is valid */
+		if (!ISABSPATH(pwd)
+		    || stat(pwd, &s_pwd) < 0 || stat(".", &s_dot) < 0
+		    || s_pwd.st_dev != s_dot.st_dev
+		    || s_pwd.st_ino != s_dot.st_ino)
+			pwdx = (char *) 0;
+		set_current_wd(pwdx);
+		if (current_wd[0])
+			simplify_path(current_wd);
+		/* Only set pwd if we know where we are or if it had a
+		 * bogus value
+		 */
+		if (current_wd[0] || pwd != null)
+			setstr(pwd_v, current_wd);
+	}
 	setint(global("PPID"), (long) getppid());
 #ifdef KSH
 	setint(global("RANDOM"), (long) time((time_t *)0));
@@ -225,7 +227,10 @@ main(argc, argv, envp)
 	}
 
 	if (geteuid() == 0)
-		setstr(global("PS1"), "# ");
+		safe_prompt = "# ";
+	else
+		safe_prompt = "$ ";
+	setstr(global("PS1"), safe_prompt);
 
 	/* Set this before parsing arguments */
 	Flag(FPRIVILEGED) = getuid() != geteuid() || getgid() != getegid();
@@ -270,13 +275,11 @@ main(argc, argv, envp)
 	}
 
 	/* This bizarreness is mandated by POSIX */
-	if (fstat(0, &s1) >= 0 && S_ISCHR(s1.st_mode)) {
-		int flags = fcntl(0, F_GETFL, 0);
+	{
+		struct stat s_stdin;
 
-		if (flags >= 0 & (flags & O_NONBLOCK)) {
-			flags &= O_NONBLOCK;
-			fcntl(0, F_SETFL, flags);
-		}
+		if (fstat(0, &s_stdin) >= 0 && S_ISCHR(s_stdin.st_mode))
+			reset_nonblock(0);
 	}
 
 	/* initialize job control */
@@ -294,18 +297,28 @@ main(argc, argv, envp)
 	restricted = Flag(FRESTRICTED);
 	Flag(FRESTRICTED) = 0;
 
-	if (argv[0][0] == '-'
-	    || ((arg = strrchr_dirsep(argv[0])) && *++arg == '-'))
-	{
+	/* Do this before profile/$ENV so that if it causes problems in them,
+	 * user will know why things broke.
+	 */
+	if (!current_wd[0] && Flag(FTALKING))
+		warningf(FALSE, "Cannot determine current working directory");
+
+	if (Flag(FLOGIN)) {
 #ifdef OS2
-		/* Need a startup option for OS/2  (4.9 used -0 as an option)
-		 * since you don't really login to os2.  This is not yet
-		 * implemented
+		char *profile;
+
+		/* Try to find a profile - first see if $INIT has a value,
+		 * then try /etc/profile.ksh, then c:/usr/etc/profile.ksh.
 		 */
-	        arg = substitute("$INIT/profile.ksh", 0);
- 		include(*arg ? arg : "/etc/profile.ksh", 0, (char **) 0);
-		include(*arg ? arg : "c:/usr/etc/profile.ksh", 0, (char **) 0);
-		include(substitute("$HOME/profile.ksh", 0), 0, (char **) 0);
+		if (!Flag(FPRIVILEGED)
+		    && strcmp(profile = substitute("$INIT/profile.ksh", 0),
+			      "/profile.ksh"))
+			include(profile, 0, (char **) 0);
+		else if (include("/etc/profile.ksh", 0, (char **) 0) < 0)
+			include("c:/usr/etc/profile.ksh", 0, (char **) 0);
+		if (!Flag(FPRIVILEGED))
+			include(substitute("$HOME/profile.ksh", 0), 0,
+				(char **) 0);
 #else /* OS2 */
 		include("/etc/profile", 0, (char **) 0);
 		if (!Flag(FPRIVILEGED))
@@ -317,10 +330,18 @@ main(argc, argv, envp)
 	if (Flag(FPRIVILEGED))
 		include("/etc/suid_profile", 0, (char **) 0);
 	else {
+		char *env_file;
+
 		/* include $ENV */
-		arg = substitute(strval(global("ENV")), DOTILDE);
-		if (*arg != '\0')
-			include(arg, 0, (char **) 0);
+		env_file = strval(global("ENV"));
+#ifdef DEFAULT_ENV
+		/* If env isn't set, include default environment */
+		if (env_file == null)
+			env_file = DEFAULT_ENV;
+#endif /* DEFAULT_ENV */
+		env_file = substitute(env_file, DOTILDE);
+		if (*env_file != '\0')
+			include(env_file, 0, (char **) 0);
 #ifdef OS2
 		else if (Flag(FTALKING))
 			include(substitute("$HOME/kshrc.ksh", 0), 0,
@@ -446,10 +467,8 @@ shell(s, exit_atend)
 
 	newenv(E_PARSE);
 	exstat = 0;
-#ifdef JOBS
 	if (interactive)
 		really_exit = 0;
-#endif /* JOBS */
 	if ((i = setjmp(e->jbuf))) {
 		s->str = null;
 		switch (i) {
@@ -498,8 +517,8 @@ shell(s, exit_atend)
 		if (interactive)
 			j_notify();
 
-		if ((wastty = (s->type == STTY)) || s->type == SHIST) {
-			set_prompt(PS1);
+		if ((wastty = (s->type == STTY))) {
+			set_prompt(PS1, s);
 			mcheck();
 		}
 
@@ -508,14 +527,14 @@ shell(s, exit_atend)
 			if (wastty && Flag(FIGNOREEOF) && --attempts > 0) {
 				shellf("Use `exit' to leave ksh\n");
 				s->type = STTY;
-#ifdef	JOBS
-			} else if (wastty && !really_exit && j_stopped()) {
+			} else if (wastty && !really_exit
+				   && j_stopped_running())
+			{
 				really_exit = 1;
 				s->type = STTY;
-#endif	/* JOBS */
 			} else {
 				/* this for POSIX, which says EXIT traps
-				 * shell be taken in the environment
+				 * shall be taken in the environment
 				 * immediately after the last command
 				 * executed.
 				 */
@@ -528,10 +547,8 @@ shell(s, exit_atend)
 		if (t && (!Flag(FNOEXEC) || s->type == STTY))
 			exstat = execute(t, 0);
 
-#ifdef JOBS
 		if (t != NULL && t->type != TEOF && interactive && really_exit)
 			really_exit = 0;
-#endif	/* JOBS */
 
 		reclaim();
 	}
