@@ -3,6 +3,7 @@
  */
 
 #include "sh.h"
+#include <ctype.h>	/* for FILECHCONV */
 #ifdef HAVE_LIMITS_H
 # include <limits.h>
 #endif
@@ -134,6 +135,7 @@ struct option options[] = {
 	{ "ignoreeof",	  0,		OF_ANY },
 	{ "interactive",'i',	    OF_CMDLINE },
 	{ "keyword",	'k',		OF_ANY },
+	{ "login",	'l',	    OF_CMDLINE },
 	{ "markdirs",	'X',		OF_ANY },
 #ifdef JOBS
 	{ "monitor",	'm',		OF_ANY },
@@ -143,11 +145,13 @@ struct option options[] = {
 	{ "noclobber",	'C',		OF_ANY },
 	{ "noexec",	'n',		OF_ANY },
 	{ "noglob",	'f',		OF_ANY },
+	{ "nohup",	  0,		OF_ANY },
 	{ "nolog",	  0,		OF_ANY }, /* no effect */
 #ifdef	JOBS
 	{ "notify",	'b',		OF_ANY },
 #endif	/* JOBS */
 	{ "nounset",	'u',		OF_ANY },
+	{ "physical",	  0,		OF_ANY }, /* non-standard */
 	{ "posix",	  0,		OF_ANY }, /* non-standard */
 	{ "privileged",	'p',		OF_ANY },
 	{ "restricted",	'r',	    OF_CMDLINE },
@@ -316,7 +320,7 @@ parse_args(argv, what, setargsp)
 	static char cmd_opts[NELEM(options) + 3]; /* o:\0 */
 	static char set_opts[NELEM(options) + 5]; /* Ao;s\0 */
 	char *opts;
-	char *arrayname;
+	char *array;
 	Getopt go;
 	int i, optc, set, sortargs = 0, arrayset = 0;
 
@@ -335,12 +339,22 @@ parse_args(argv, what, setargsp)
 		strcpy(set_opts, "Ao;s"); /* see set_opts[] declaration */
 		p = set_opts + strlen(set_opts);
 		for (i = 0; options[i].name; i++)
-			if (options[i].c && (options[i].flags & OF_CMDLINE))
+			if (options[i].c && (options[i].flags & OF_SET))
 				*p++ = options[i].c;
 		*p = '\0';
 	}
 
-	opts = (what == OF_CMDLINE) ? cmd_opts : set_opts;
+	if (what == OF_CMDLINE) {
+		char *p;
+		/* Set FLOGIN before parsing options so user can clear
+		 * flag using +l.
+		 */
+		Flag(FLOGIN) = (argv[0][0] == '-'
+				|| ((p = strrchr_dirsep(argv[0]))
+				     && *++p == '-'));
+		opts = cmd_opts;
+	} else
+		opts = set_opts;
 	ksh_getopt_reset(&go, GF_ERROR|GF_PLUSOPT);
 	while ((optc = ksh_getopt(argv, &go, opts)) != EOF) {
 		set = (go.info & GI_PLUS) ? 0 : 1;
@@ -408,18 +422,17 @@ parse_args(argv, what, setargsp)
 					  || argv[go.optind]);
 
 	if (arrayset) {
-		arrayname = argv[go.optind++];
-		if (!arrayname) {
+		array = argv[go.optind++];
+		if (!array) {
 			bi_errorf("-A: missing array name\n");
 			return -1;
 		}
-		if (!*arrayname || *skip_varname(arrayname, FALSE)) {
-			bi_errorf("%s: is not an identifier\n",
-				arrayname);
+		if (!*array || *skip_varname(array, FALSE)) {
+			bi_errorf("%s: is not an identifier\n", array);
 			return -1;
 		}
 	} else
-		arrayname = (char *) 0;	/* keep gcc happy */
+		array = (char *) 0;	/* keep gcc happy */
 	if (sortargs) {
 		for (i = go.optind; argv[i]; i++)
 			;
@@ -427,7 +440,7 @@ parse_args(argv, what, setargsp)
 			xstrcmp);
 	}
 	if (arrayset) {
-		set_array(arrayname, arrayset, argv + go.optind);
+		set_array(array, arrayset, argv + go.optind);
 		for (; argv[go.optind]; go.optind++)
 			;
 	}
@@ -446,7 +459,7 @@ getn(as, ai)
 	int sawdigit = 0;
 
 	s = as;
-	if (*s == '-')
+	if (*s == '-' || *s == '+')
 		s++;
 	for (n = 0; digit(*s); s++, sawdigit = 1)
 		n = n * 10 + (*s - '0');
@@ -480,15 +493,20 @@ bi_getn(as, ai)
  */
 
 int
-gmatch(s, p)
+gmatch(s, p, isfile)
 	register char *s, *p;
+	int isfile;
 {
-	register int sc, pc;
+	register unsigned int sc, pc;
 
 	if (s == NULL || p == NULL)
 		return 0;
 	while ((pc = *p++) != 0) {
-		sc = FILECHCONV(*s++);
+		sc = *s++;
+		if (isfile) {
+			sc = FILECHCONV(sc);
+			pc = FILECHCONV(pc);
+		}
 		if (pc == MAGIC)
 			switch (*p++) {
 			  case '[':
@@ -504,7 +522,7 @@ gmatch(s, p)
 			  case '*':
 				s--;
 				do {
-					if (*p == '\0' || gmatch(s, p))
+					if (*p == '\0' || gmatch(s, p, isfile))
 						return (1);
 				} while (*s++ != '\0');
 				return (0);
@@ -697,7 +715,7 @@ ksh_getopt_reset(go, flags)
  *	- ',' is like ':' in options, except the argument always immediately
  *	  follows the option character (optarg is set to the null string if
  *	  the option is missing).
- *	  Used for 'read -u2' and 'print -u2'.
+ *	  Used for 'read -u2', 'print -u2' and fc -40.
  *	- '#' is like ':' in options, expect that the argument is optional
  *	  and must start with a digit.  If the argument doesn't start with a
  *	  digit, it is assumed to be missing and normal option processing
@@ -717,18 +735,14 @@ ksh_getopt(argv, go, options)
 	char c;
 	char *o;
 
-	/* protect against multiple calls which could cause optind or p
-	 * to go out of range.  Also prevents skipping --.
-	 */
-	if (go->info & GI_DONE)
-		return EOF;
 	if (go->p == 0 || (c = argv[go->optind - 1][go->p]) == '\0') {
 		char *arg = argv[go->optind], flag = arg ? *arg : '\0';
 
 		go->p = 1;
 		if (flag == '-' && arg[1] == '-' && arg[2] == '\0') {
 			go->optind++;
-			go->info |= GI_DONE | GI_MINUSMINUS;
+			go->p = 0;
+			go->info |= GI_MINUSMINUS;
 			return EOF;
 		}
 		if (arg == (char *) 0
@@ -737,7 +751,7 @@ ksh_getopt(argv, go, options)
 			    || flag != '+'))
 		    || (c = arg[1]) == '\0')
 		{
-			go->info |= GI_DONE;
+			go->p = 0;
 			return EOF;
 		}
 		go->optind++;
@@ -747,13 +761,13 @@ ksh_getopt(argv, go, options)
 	if (c == '?' || c == ':' || c == ';' || c == ',' || c == '#'
 	    || !(o = strchr(options, c)))
 	{
-		go->info |= GI_DONE;
 		if (options[0] == ':') {
 			go->buf[0] = c;
 			go->optarg = go->buf;
 		} else {
-			warningf(TRUE, "%s: -%c: bad option\n",
-				argv[0], c);
+			warningf(TRUE, "%s%s-%c: bad option",
+				(go->flags & GI_NONAME) ? "" : argv[0],
+				(go->flags & GI_NONAME) ? "" : ": ", c);
 			if (go->flags & GF_ERROR)
 				bi_errorf((char *) 0);
 		}
@@ -772,14 +786,14 @@ ksh_getopt(argv, go, options)
 		else if (*o == ';')
 			go->optarg = (char *) 0;
 		else {
-			go->info |= GI_DONE;
 			if (options[0] == ':') {
 				go->buf[0] = c;
 				go->optarg = go->buf;
 				return ':';
 			}
-			warningf(TRUE, "%s: -`%c' requires argument\n",
-				argv[0], c);
+			warningf(TRUE, "%s%s-`%c' requires argument",
+				(go->flags & GI_NONAME) ? "" : argv[0],
+				(go->flags & GI_NONAME) ? "" : ": ", c);
 			if (go->flags & GF_ERROR)
 				bi_errorf((char *) 0);
 			return '?';
@@ -949,3 +963,134 @@ str_zcpy(dst, src, dsize)
 	}
 	return dst;
 }
+
+/* Like read(2), but if read fails due to non-blocking flag, resets flag
+ * and restarts read.
+ */
+int
+blocking_read(fd, buf, nbytes)
+	int fd;
+	char *buf;
+	int nbytes;
+{
+	int ret;
+	int tried_reset = 0;
+
+	while ((ret = read(fd, buf, nbytes)) < 0) {
+		if (!tried_reset && (errno == EAGAIN
+#ifdef EWOULDBLOCK
+				     || errno == EWOULDBLOCK
+#endif /* EWOULDBLOCK */
+				    ))
+		{
+			int oerrno = errno;
+			if (reset_nonblock(fd) > 0) {
+				tried_reset = 1;
+				continue;
+			}
+			errno = oerrno;
+		}
+		break;
+	}
+	return ret;
+}
+
+/* Reset the non-blocking flag on the specified file descriptor.
+ * Returns -1 if there was an error, 0 if non-blocking wasn't set,
+ * 1 if it was.
+ */
+int
+reset_nonblock(fd)
+	int fd;
+{
+	int flags;
+	int blocking_flags;
+
+	if ((flags = fcntl(fd, F_GETFL, 0)) < 0)
+		return -1;
+	/* With luck, the C compiler will reduce this to a constant */
+	blocking_flags = 0;
+#ifdef O_NONBLOCK
+	blocking_flags |= O_NONBLOCK;
+#endif /* O_NONBLOCK */
+#ifdef O_NDELAY
+	blocking_flags |= O_NDELAY;
+#else /* O_NDELAY */
+# ifndef O_NONBLOCK
+	blocking_flags |= FNDELAY; /* hope this exists... */
+# endif /* O_NONBLOCK */
+#endif /* O_NDELAY */
+	if (!(flags & blocking_flags))
+		return 0;
+	flags &= ~blocking_flags;
+	if (fcntl(fd, F_SETFL, flags) < 0)
+		return -1;
+	return 1;
+}
+
+
+#ifdef HAVE_SYS_PARAM_H
+# include <sys/param.h>
+#endif /* HAVE_SYS_PARAM_H */
+#ifndef MAXPATHLEN
+# define MAXPATHLEN PATH
+#endif /* MAXPATHLEN */
+
+/* Like getcwd(), except bsize is ignored if buf is 0 (MAXPATHLEN is used) */
+char *
+ksh_get_wd(buf, bsize)
+	char *buf;
+	int bsize;
+{
+#ifdef HAVE_GETWD
+	extern char *getwd ARGS((char *));
+	char *b;
+	int len;
+
+	if (buf && bsize > MAXPATHLEN)
+		b = buf;
+	else
+		b = alloc(MAXPATHLEN + 1, ATEMP);
+	if (!getwd(b)) {
+		errno = EACCES;
+		if (b != buf)
+			afree(b, ATEMP);
+		return (char *) 0;
+	}
+	len = strlen(b) + 1;
+	if (!buf)
+		b = aresize(b, len, ATEMP);
+	else if (buf != b) {
+		if (len > bsize) {
+			errno = ERANGE;
+			return (char *) 0;
+		}
+		memcpy(buf, b, len);
+		afree(b, ATEMP);
+		b = buf;
+	}
+
+	return b;
+#else /* HAVE_GETWD */
+	char *b;
+	char *ret;
+
+	/* Assume getcwd() available */
+	if (!buf) {
+		bsize = MAXPATHLEN;
+		b = alloc(MAXPATHLEN + 1, ATEMP);
+	} else
+		b = buf;
+
+	ret = getcwd(b, bsize);
+
+	if (!buf) {
+		if (ret)
+			ret = aresize(b, strlen(b) + 1, ATEMP);
+		else
+			afree(b, ATEMP);
+	}
+
+	return ret;
+#endif /* HAVE_GETWD */
+}
