@@ -54,6 +54,7 @@ extern int chdir ARGS((const char *));
 extern int kill ARGS((pid_t, int));
 extern char *getcwd();	/* no ARGS here - differs on different machines */
 extern int geteuid ARGS((void));
+extern int readlink ARGS((const char *, char *, int));
 extern int getegid ARGS((void));
 extern int getpid ARGS((void));
 extern int getppid ARGS((void));
@@ -67,10 +68,10 @@ extern int setpgid ARGS((pid_t, pid_t));
 extern int getpgrp ARGS((pid_t));
 extern int setpgrp ARGS((pid_t, pid_t));
 # endif /* BSD_PGRP */
-# ifdef SVR3_PGRP
+# ifdef SYSV_PGRP
 extern int getpgrp ARGS((void));
 extern int setpgrp ARGS((void));
-# endif /* SVR3_PGRP */
+# endif /* SYSV_PGRP */
 #endif /* HAVE_UNISTD_H */
 
 #ifdef HAVE_STRING_H
@@ -122,21 +123,26 @@ extern int errno;
 #endif /* HAVE_FCNTL_H */
 #ifndef O_ACCMODE
 # define O_ACCMODE	(O_RDONLY|O_WRONLY|O_RDWR)
-#endif /* O_ACCMODE */
-#ifndef O_NONBLOCK
-# ifdef O_NDELAY
-#  define O_NONBLOCK	O_NDELAY
-# else /* O_NDELAY */
-#  define O_NONBLOCK	FNDELAY		/* hope this exists... */
-# endif /* O_NDELAY */
-#endif /* O_NONBLOCK */
+#endif /* !O_ACCMODE */
 
 #ifndef F_OK 	/* access() arguments */
 # define F_OK 0
 # define X_OK 1
 # define W_OK 2
 # define R_OK 4
-#endif /* F_OK */
+#endif /* !F_OK */
+
+#ifndef SEEK_SET
+# ifdef L_SET
+#  define SEEK_SET L_SET
+#  define SEEK_CUR L_INCR
+#  define SEEK_END L_XTND
+# else /* L_SET */
+#  define SEEK_SET 0
+#  define SEEK_CUR 1
+#  define SEEK_END 2
+# endif /* L_SET */
+#endif /* !SEEK_SET */
 
 /* Some machines (eg, FreeBSD 1.1.5) define CLK_TCK in limits.h
  * (ksh_limval.h assumes limits has been included, if available)
@@ -196,24 +202,32 @@ extern int errno;
 #ifdef OS2
 # define ksh_execve(p, av, ev)	_execve(p, av, ev)
 #else /* OS2 */
-# if defined(isc386) && defined(_POSIX_SOURCE)
+# if defined(OS_ISC) && defined(_POSIX_SOURCE)
 /* Kludge for ISC 3.2 (and other versions?) so programs will run correctly.  */
 #  define ksh_execve(p, av, ev) do { \
 					__setostype(0); \
 					execve(p, av, ev); \
 					__setostype(1); \
 				} while (0)
-# else /* isc386 && _POSIX */
+# else /* OS_ISC && _POSIX */
 #  define ksh_execve(p, av, ev)	execve(p, av, ev)
-# endif /* isc386 && _POSIX */
+# endif /* OS_ISC && _POSIX */
 #endif /* OS2 */
 
-#ifdef OS2
-extern int getdup ARGS((int fd, int base));
-# define ksh_dupbase(fd, base) getdup(fd, base)
-#else /* OS2 */
-# define ksh_dupbase(fd, base) fcntl(fd, F_DUPFD, base)
-#endif /* OS2 */
+/* this is a hang-over from older versions of the os2 port */
+#define ksh_dupbase(fd, base) fcntl(fd, F_DUPFD, base)
+
+/* Find a 32 bit integer type (or die) - SIZEOF_* defined by autoconf
+ * (assumes 8 bit byte, but I'm not concerned) */
+#if SIZEOF_INT == 4
+# define INT32	int
+#else /* SIZEOF_INT */
+# if SIZEOF_LONG == 4
+#  define INT32	long
+# else /* SIZEOF_LONG */
+   #error cannot find 32 bit type...
+# endif /* SIZEOF_LONG */
+#endif /* SIZEOF_INT */
 
 /* end of common headers */
 
@@ -244,15 +258,31 @@ extern int getdup ARGS((int fd, int base));
 # endif /* OS2 */
 #endif
 
-/* These defines aren't used everywhere yet, but it is being worked on */
+/* ISABSPATH() means path is fully and completely specified,
+ * ISROOTEDPATH() means a .. as the first component is a no-op,
+ * ISRELPATH() means $PWD can be tacked on to get an absolute path.
+ *
+ * OS	Path		ISABSPATH	ISROOTEDPATH	ISRELPATH
+ * unix	/foo		yes		yes		no
+ * unix	foo		no		no		yes
+ * unix	../foo		no		no		yes
+ * os2	a:/foo		yes		yes		no
+ * os2	a:foo		no		no		no
+ * os2	/foo		no		yes		no
+ * os2	foo		no		no		yes
+ * os2	../foo		no		no		yes
+ */
 #ifdef OS2
 # define PATHSEP        ';'
 # define DIRSEP         '\\'
 # define DIRSEPSTR      "\\"
 # define ISDIRSEP(c)    ((c) == '\\' || (c) == '/')
-# define FILECHCONV(c)	tolower(c)
+# define ISABSPATH(s)	(((s)[0] && (s)[1] == ':' && ISDIRSEP((s)[2])))
+# define ISROOTEDPATH(s) (ISDIRSEP((s)[0]) || ISABSPATH(s))
+# define ISRELPATH(s)	(!(s)[0] || ((s)[1] != ':' && !ISDIRSEP((s)[0])))
+# define FILECHCONV(c)	(isupper(c) ? tolower(c) : c)
 # define FILECMP(s1, s2) stricmp(s1, s2)
-# define FILENCMP(s1, s2, n) stricmp(s1, s2, n)
+# define FILENCMP(s1, s2, n) strnicmp(s1, s2, n)
 extern char *strchr_dirsep(char *path);
 extern char *strrchr_dirsep(char *path);
 # define chdir          _chdir2
@@ -262,6 +292,9 @@ extern char *strrchr_dirsep(char *path);
 # define DIRSEP         '/'
 # define DIRSEPSTR      "/"
 # define ISDIRSEP(c)    ((c) == '/')
+# define ISABSPATH(s)	ISDIRSEP((s)[0])
+# define ISROOTEDPATH(s) ISABSPATH(s)
+# define ISRELPATH(s)	(!ISABSPATH(s))
 # define FILECHCONV(c)	c
 # define FILECMP(s1, s2) strcmp(s1, s2)
 # define FILENCMP(s1, s2, n) strncmp(s1, s2, n)
@@ -296,6 +329,7 @@ EXTERN	pid_t	kshpid;		/* $$, shell pid */
 EXTERN	pid_t	procpid;	/* pid of executing process */
 EXTERN	int	exstat;		/* exit status */
 EXTERN	int	subst_exstat;	/* exit status of last $(..)/`..` */
+EXTERN	char	*safe_prompt;	/* safe prompt if PS1 substitution fails */
 
 
 /*
@@ -358,6 +392,7 @@ EXTERN	struct env {
 #define	LBREAK	6		/* break statement */
 #define	LCONTIN	7		/* continue statement */
 #define LSHELL	8		/* return to interactive shell() */
+#define LAEXPR	9		/* error in arithmetic expression */
 
 
 /* option processing */
@@ -374,7 +409,7 @@ struct option {
 extern struct option options[];
 
 /*
- * flags (the order of these enums MUST match the order in options[])
+ * flags (the order of these enums MUST match the order in misc.c(options[]))
  */
 enum sh_flag {
 	FEXPORT = 0,	/* -a: export all */
@@ -393,16 +428,19 @@ enum sh_flag {
 	FIGNOREEOF,	/* eof does not exit */
 	FTALKING,	/* -i: interactive */
 	FKEYWORD,	/* -k: name=value anywere */
+	FLOGIN,		/* -l: a login shell */
 	FMARKDIRS,	/* mark dirs with / in file name completion */
 	FMONITOR,	/* -m: job control monitoring */
 	FNOCLOBBER,	/* -C: don't overwrite existing files */
 	FNOEXEC,	/* -n: don't execute any commands */
 	FNOGLOB,	/* -f: don't do file globbing */
+	FNOHUP,		/* -H: don't kill running jobs when login shell exits */
 	FNOLOG,		/* don't save functions in history (ignored) */
 #ifdef	JOBS
 	FNOTIFY,	/* -b: asynchronous job completion notification */
 #endif
 	FNOUNSET,	/* -u: using an unset var is an error */
+	FPHYSICAL,	/* -o physical: don't do logical cd's/pwd's */
 	FPOSIX,		/* -o posix: be posixly correct */
 	FPRIVILEGED,	/* -p: use suid_profile */
 	FRESTRICTED,	/* -r: restricted shell */
@@ -539,7 +577,7 @@ EXTERN int ifs0 _I_(' ');	/* for "$*" */
 #define GF_PLUSOPT	BIT(1)	/* allow +c as an option */
 
 /* Values for Getopt.info */
-#define GI_DONE		BIT(0)	/* set when at end of options or option error */
+#define GI_NONAME	BIT(0)	/* don't print argv[0] in errors */
 #define GI_MINUS	BIT(1)	/* an option started with -... */
 #define GI_PLUS		BIT(2)	/* an option started with +... */
 #define GI_MINUSMINUS	BIT(3)	/* arguments were ended with -- */
@@ -565,13 +603,24 @@ struct coproc {
 };
 EXTERN struct coproc coproc;
 
+extern char ksh_version[];
 
 /* name of called builtin function (used by error functions) */
 EXTERN char	*builtin_argv0;
 EXTERN int	builtin_flag;	/* flags of called builtin (SPEC_BI, etc.) */
 
+/* current working directory, and size of memory allocated for same */
+EXTERN char	*current_wd;
+EXTERN int	current_wd_size;
 
 #ifdef EDIT
+/* Minimium required space to work with on a line - if the prompt leaves less
+ * space than this on a line, the prompt is truncated.
+ */
+# define MIN_EDIT_SPACE	7
+/* Minimium allowed value for x_cols: 2 for prompt, 3 for " < " at end of line
+ */
+# define MIN_COLS	(2 + MIN_EDIT_SPACE + 3)
 EXTERN	int	x_cols _I_(80);	/* tty columns */
 #else
 # define x_cols 80		/* for pr_menu(exec.c) */
