@@ -410,7 +410,8 @@ shf_fillbuf(shf)
 
 	shf->rp = shf->buf;
 	while (1) {
-		shf->rnleft = read(shf->fd, shf->buf, shf->rbsize);
+		shf->rnleft = blocking_read(shf->fd, (char *) shf->buf,
+					    shf->rbsize);
 		if (shf->rnleft < 0 && errno == EINTR
 		    && !(shf->flags & SHF_INTERRUPT))
 			continue;
@@ -429,6 +430,54 @@ shf_fillbuf(shf)
 	return 0;
 }
 
+/* Seek to a new position in the file.  If writing, flushes the buffer
+ * first.  If reading, optimizes small relative seeks that stay inside the
+ * buffer.  Returns 0 for success, EOF otherwise.
+ */
+int
+shf_seek(shf, where, from)
+	struct shf *shf;
+	off_t where;
+	int from;
+{
+	if (shf->fd < 0) {
+		errno = EINVAL;
+		return EOF;
+	}
+
+	if (shf->flags & SHF_ERROR) {
+		errno = shf->errno_;
+		return EOF;
+	}
+
+	if ((shf->flags & SHF_WRITING) && shf_emptybuf(shf, EB_READSW) == EOF)
+		return EOF;
+
+	if (shf->flags & SHF_READING) {
+		if (from == SEEK_CUR &&
+				(where < 0 ?
+					-where >= shf->rbsize - shf->rnleft :
+					where < shf->rnleft)) {
+			shf->rnleft -= where;
+			shf->rp += where;
+			return 0;
+		}
+		shf->rnleft = 0;
+		shf->rp = shf->buf;
+	}
+
+	shf->flags &= ~(SHF_EOF | SHF_READING | SHF_WRITING);
+
+	if (lseek(shf->fd, where, from) < 0) {
+		shf->errno_ = errno;
+		shf->flags |= SHF_ERROR;
+		return EOF;
+	}
+
+	return 0;
+}
+
+
 /* Read a buffer from shf.  Returns the number of bytes read into buf,
  * if no bytes were read, returns 0 if end of file was seen, EOF if
  * a read error occurred.
@@ -606,6 +655,20 @@ shf_putchar(c, shf)
 	return c;
 }
 
+/* Write a string.  Returns the length of the string if successful, EOF if
+ * the string could not be written.
+ */
+int
+shf_puts(s, shf)
+	char *s;
+	struct shf *shf;
+{
+	if (!s)
+		return EOF;
+
+	return shf_write(s, strlen(s), shf);
+}
+
 /* Write a buffer.  Returns nbytes if successful, EOF if there is an error. */
 int
 shf_write(buf, nbytes, shf)
@@ -704,7 +767,7 @@ shf_snprintf(buf, bsize, fmt, va_alist)
 	int n;
 
 	if (!buf || bsize <= 0)
-		internal_errorf(1, "shf_sprintf: buf %lx, bsize %d",
+		internal_errorf(1, "shf_snprintf: buf %lx, bsize %d",
 			(long) buf, bsize);
 
 	shf_sopen(buf, bsize, SHF_WR, &shf);
@@ -715,6 +778,26 @@ shf_snprintf(buf, bsize, fmt, va_alist)
 	return n;
 }
 
+char *
+#ifdef HAVE_PROTOTYPES
+shf_smprintf(char *fmt, ...)
+#else
+shf_smprintf(fmt, va_alist)
+	char *fmt;
+	va_dcl
+#endif
+{
+	struct shf shf;
+	va_list args;
+	int n;
+
+	shf_sopen((char *) 0, 0, SHF_WR|SHF_DYNAMIC, &shf);
+	SH_VA_START(args, fmt);
+	n = shf_vfprintf(&shf, fmt, args);
+	va_end(args);
+	return shf_sclose(&shf); /* null terminates */
+}
+
 #undef FP  			/* if you want floating point stuff */
 
 #define BUF_SIZE	128
